#!/usr/bin/python
"""
A* Demo
by Kris Schnee.
Demonstrates the "A* Algorithm," a standard AI technique used for finding
an efficient path from one place to another. (It could also be used for
more abstract forms of goal-seeking, or take into account factors other
than ease of movement.) It assumes complete knowledge of the terrain
covered, and the cost of movement between squares. It also uses an optimistic
method for estimating the cost from a point to the goal, but you can change
this method.
Credit goes to <http://en.wikipedia.org/wiki/A*> for describing the algorithm.
I also recycled a picture of a shrub from some free graphics at lostgarden.com.
"""

##___________________________________
##          ~Header~
##___________________________________

__author__ = "Kris Schnee"
__version__ = "2008.7.21"
__license__ = "Public Domain"


##___________________________________
##      ~Imported Modules~
##___________________________________

## Standard
import random

## Third-party
import pygame

## Mine

##___________________________________
##            ~Constants~
##___________________________________

WORLD_SIZE_X = 8
WORLD_SIZE_Y = 6

MAX_TRIES = 100

pygame.init()
screen = pygame.display.set_mode((800,600))
font = pygame.font.Font(None,32)

TILES = {"grass":pygame.image.load("grass.png").convert(),
         "water":pygame.image.load("water.png").convert(),
         "forest":pygame.image.load("forest.png").convert(),
         }
MARKS = {"path":(255,255,255),
         "start":(255,0,0),
         "tried":(127,127,127),
         "goal":(0,255,0),
         }
COSTS = {"grass":1,
         "forest":2,
         "water":9,
         }
AVERAGE_MOVE_COST = int(sum(COSTS.values())/len(COSTS))
ASSUME_CONSTANT_TERRAIN = False ## If True, use the above value for cost.
ASSUMED_MOVE_COST = min(COSTS.values()) ## An optimistic estimate.

##___________________________________
##            ~Classes~
##___________________________________

class World:
    def __init__(self,layout="random"):
        self.land = []

        if layout == "seeded":
            random.seed("Skunktaur") ## Makes a nice example.
            layout = "random"
        if layout == "random":
            for x in range(WORLD_SIZE_X):
                self.land.append([])
                for y in range(WORLD_SIZE_Y):
                    roll = random.randint(0,99)
                    if roll < 29:
                        terrain = "water"
                    elif roll < 40:
                        terrain = "forest"
                    else:
                        terrain = "grass"
                    tile = {"terrain":terrain,"mark":None}
                    self.land[-1].append(tile)

        elif layout == "wall":
            for x in range(WORLD_SIZE_X):
                self.land.append([])
                for y in range(WORLD_SIZE_Y):
                    tile = {"terrain":"grass","mark":None}
                    self.land[-1].append(tile)

            self.land[3][1]["terrain"] = "water"
            self.land[3][2]["terrain"] = "water"
            self.land[3][3]["terrain"] = "water"
            self.land[3][4]["terrain"] = "water"
            self.land[4][2]["terrain"] = "forest"
##            self.land[2][0]["terrain"] = "forest"

        elif layout == "cave":
            for x in range(WORLD_SIZE_X):
                self.land.append([])
                for y in range(WORLD_SIZE_Y):
                    tile = {"terrain":"grass","mark":None}
                    self.land[-1].append(tile)
            cave = [(3,1),(3,2),(3,3),(3,4),
                    (2,1),(1,1),(2,4),(1,4),]
            for x, y in cave:
                self.land[x][y]["terrain"] = "water"


##        self.land[4][3]["terrain"] = "forest"
##        self.land[4][4]["terrain"] = "forest"

    def Draw(self,seeker,text_markers=False,open_markers=False):

        ## Mark tiles in the "open" list.
        if open_markers:
            for note in seeker.open:
                x,y = note
                self.land[x][y]["mark"] = "tried"

        ## Mark the selected path if one was found.
        if seeker.chosen_path:
            for x, y in seeker.chosen_path:
                self.land[x][y]["mark"] = "path"

            x,y = seeker.start
            self.land[x][y]["mark"] = "start"
            x,y = seeker.goal
            self.land[x][y]["mark"] = "goal"

        ## Draw the tiles and any markers on them.
        for x in range(WORLD_SIZE_X):
            for y in range(WORLD_SIZE_Y):
                tile = self.land[x][y]

                img = TILES[tile["terrain"]]
                screen.blit(img,(x*100,y*100))

                mark = tile["mark"]
                if mark:
                    c = MARKS[mark]
                    pygame.draw.rect(screen,c,(x*100+20,y*100+20,60,60))

        ## Add numeric labels of cost-to-reach and back-tracing.
        if text_markers:
            for key in seeker.knowledge:
                value = seeker.knowledge[key]
                cost = str(value["cost_to_here"])
                if len(cost) > 3:
                    cost = "Lots"
                text = font.render(cost,1,(0,0,255))
                screen.blit(text,(key[0]*100+40,key[1]*100+15))

                text = font.render(str(value["from"]),1,(0,0,255))
                screen.blit(text,(key[0]*100+20,key[1]*100+40))

        pygame.display.update()




class AStar:
    def __init__(self):
        global w
        self.world = w
        self.knowledge = {}
        self.open = []
        self.closed = []
        self.chosen_path = []
        self.start = None
        self.goal = None

    def Plan(self,start,end):
        self.start = start
        d = Dist(start,end)
        tile = {"coords":start,
                "cost_to_here":0,
                "cost_from_here":d,
                "total_cost":d,
                "from":None,
                "start":True}
        self.knowledge[start] = tile
        self.open.append(start)
        self.goal = end
        self.tries = 0
        self.max_tries = MAX_TRIES

        while self.open and self.tries < self.max_tries:
            self.tries += 1

            self.open.sort( lambda a,b: cmp(self.knowledge[b]["total_cost"],self.knowledge[a]["total_cost"]) )
            coords = self.open.pop()
            self.closed.append(coords)

            knowledge = self.knowledge[coords]
            x, y = coords

            if coords == self.goal:
                alltiles = self.closed + self.open
                print "*** Reached the goal. ***"

                self.chosen_path = [ coords ]

                n = 0
                while coords != self.start:
                    n += 1
                    if n > 50:
                        raise "oops"

                    knowledge = self.knowledge[coords]
##                    print str(coords) + " comes from "+str(knowledge["from"])
                    coords = knowledge["from"]
                    self.chosen_path.append(coords)

                print "Path (end to beginning): "+str(self.chosen_path)
                return

            neighbors = []
            if x > 0:
                neighbors.append((x-1,y))
            if x < WORLD_SIZE_X - 1:
                neighbors.append((x+1,y))
            if y > 0:
                neighbors.append((x,y-1))
            if y < WORLD_SIZE_Y - 1:
                neighbors.append((x,y+1))            

            for n in neighbors:
                cost_from_neighbor = Dist(n,self.goal)
                if ASSUME_CONSTANT_TERRAIN:
                    step_cost = AVERAGE_MOVE_COST ## Assume constant terrain.
                else:
                    terrain = self.world.land[n[0]][n[1]]["terrain"]
                    step_cost = COSTS[terrain]
                    cost_from_neighbor *= ASSUMED_MOVE_COST
                
                cost_to_neighbor = knowledge["cost_to_here"] + step_cost

                if n in self.closed:
                    continue

                if not n in self.open:
                    ## This is not in the open list.
                    tile = {"coords":n,
                            "cost_to_here":cost_to_neighbor,
                            "cost_from_here":cost_from_neighbor,
                            "total_cost":cost_to_neighbor+cost_from_neighbor,
                            "from":coords,
                            }
                    self.knowledge[n] = tile
                    self.open.append(n)

                else:
                    neighbor_knowledge = self.knowledge[n]
                    existing_cost_to_neighbor = neighbor_knowledge["cost_to_here"]
                    if cost_to_neighbor < existing_cost_to_neighbor:
                        ## Found a cheaper route to here.
                        neighbor_knowledge["from"] = coords
                        neighbor_knowledge["cost_to_here"] = cost_to_neighbor
                        neighbor_knowledge["total_cost"] = cost_to_neighbor+cost_from_neighbor

        print "I wasn't able to find a path within "+str(MAX_TRIES)+" attempts. Bailing."

##___________________________________
##            ~Functions~
##___________________________________

def Dist(p1,p2,ortho=True):
    if ortho:
        return abs(p1[0]-p2[0]) + abs(p1[1]-p2[1])


##__________________________________
##            ~Autorun~
##__________________________________

if __name__ == "__main__":
    import pygame

##    w = World(layout="wall")
##    w = World(layout="random")
##    w = World(layout="cave")
    w = World(layout="seeded")

    seeker = AStar()
    seeker.Plan((0,1),(6,4))
##    seeker.Plan((1,2),(7,3))
##    w.Draw(seeker,True,True)
    w.Draw(seeker,False,False)

